<?php

/**
 * @file
 * Defines the classes necessary for a Grouping type question
 */

/**
 * Extension of QuizQuestion.
 */
class GroupingQuestion extends QuizQuestion {

  /**
   * Implements saveNodeProperties().
   * @see QuizQuestion#saveNodeProperties($is_new)
   */
  public function saveNodeProperties($is_new = FALSE) {

    $result = _grouping_question_save($this->node, $is_new);

    // We fetch ids for the existing groups belonging to this question
    // We need to figure out if an existing group has been changed or deleted.
    $sql = 'SELECT id FROM {quiz_grouping_question_group}
      WHERE question_nid = %d AND question_vid = %d';
    $res = db_query($sql, $this->node->nid, $this->node->vid);

    // We start by assuming that all existing alternatives needs to be deleted.
    $ids_to_delete = array();
    while ($res_o = db_fetch_object($res)) {
      $ids_to_delete[] = $res_o->id;
    }

    for ($i = 1; isset($this->node->groups[$i]); $i++) {
      $short = $this->node->groups[$i];
      if (drupal_strlen($short['name']) > 0) {
        if (!is_numeric($short['id'])) {
          // New group.
          $this->insertStatement($i);
        }
        else {
          // Existing group. Check if it is a new revision.
          $key = array_search($short['id'], $ids_to_delete);
          if ($key === FALSE) {
            // New revision.
            $this->insertStatement($i);
          }
          else {
            $this->updateStatement($i);
            // Make sure this alternative isn't deleted.
            $ids_to_delete[$key] = FALSE;
          }
        }
      }
    }
    foreach ($ids_to_delete as $id_to_delete) {
      if ($id_to_delete) {
        db_query('DELETE FROM {quiz_grouping_question_group} WHERE id = %d', $id_to_delete);
      }
    }
  }

  /**
   * Helper function. Saves new group
   *
   * @param int $i
   *   The numerical index of the group to save
   */
  public function insertStatement($i) {
    $sql = 'INSERT INTO {quiz_grouping_question_group}
      (name, members, feedback, question_nid, question_vid)
      VALUES(\'%s\', \'%s\', \'%s\', %d, %d)';
    $short = $this->node->groups[$i];
    db_query($sql, $short['name'], _grouping_question_clean_csv($short['members']), $short['feedback'], $this->node->nid, $this->node->vid);
  }

  /**
   * Helper function. Update existing group
   *
   * @param int $i
   *   The numerical index of the group to save
   */
  public function updateStatement($i) {
    $sql = 'UPDATE {quiz_grouping_question_group}
      SET name = \'%s\', members = \'%s\', feedback = \'%s\'
      WHERE id = %d AND question_nid = %d AND question_vid = %d';
    $short = $this->node->groups[$i];
    db_query($sql, $short['name'], _grouping_question_clean_csv($short['members']), $short['feedback'], $short['id'], $this->node->nid, $this->node->vid);
  }

  /**
   * Implements validateNode().
   * @see QuizQuestion#validateNode($form)
   */
  public function validateNode(array &$form) {
    // TODO: validate that each member only exists in one place.
  }

  /**
   * Implements delete().
   * @see QuizQuestion#delete()
   */
  public function delete($only_this_version = FALSE) {
    $delete_properties = 'DELETE FROM {quiz_grouping_question_properties} WHERE nid = %d';
    $delete_targets = 'DELETE FROM {quiz_grouping_question_group} WHERE question_nid = %d';
    $delete_multi_results = '
      DELETE FROM {quiz_grouping_question_user_answer_multi}
      USING {quiz_grouping_question_user_answer_multi}
      INNER JOIN {quiz_grouping_question_user_answers}
      ON {quiz_grouping_question_user_answer_multi}.user_answer_id = {quiz_grouping_question_user_answers}.id
      WHERE {quiz_grouping_question_user_answers}.question_nid = %d' . ($only_this_version ? ' AND question_vid = %d' : '');
    $delete_results = 'DELETE FROM {quiz_grouping_question_user_answers} WHERE question_nid = %d';
    if ($only_this_version) {
      $delete_properties .= ' AND vid = %d';
      $delete_targets .= ' AND question_vid = %d';
      $delete_results .= ' AND question_vid = %d';
    }
    db_query($delete_properties, $this->node->nid, $this->node->vid);
    db_query($delete_targets, $this->node->nid, $this->node->vid);
    db_query($delete_multi_results, $this->node->nid, $this->node->vid);
    db_query($delete_results, $this->node->nid, $this->node->vid);

    parent::delete($only_this_version);
  }

  /**
   * Implements getNodeProperties().
   * @see QuizQuestion#getNodeProperties()
   */
  public function getNodeProperties() {
    if (isset($this->nodeProperties)) {
      return $this->nodeProperties;
    }
    $props = parent::getNodeProperties();

    // Load the properties.
    $sql = 'SELECT * FROM {quiz_grouping_question_properties} WHERE nid = %d AND vid = %d';
    $res = db_query($sql, $this->node->nid, $this->node->vid);
    $res_a = db_fetch_array($res);
    if (is_array($res_a)) {
      $props = array_merge($props, $res_a);
    }

    // Load the groups.
    $sql = 'SELECT *
      FROM {quiz_grouping_question_group}
      WHERE question_nid = %d AND question_vid = %d
      ORDER BY id';
    $res = db_query($sql, $this->node->nid, $this->node->vid);
    $props['groups'] = array();

    $i = 1;
    while ($res_arr = db_fetch_array($res)) {
      $props['groups'][$i] = $res_arr;
      $i++;
    }
    $this->nodeProperties = $props;
    return $props;
  }

  /**
   * Implements getNodeView().
   * @see QuizQuestion#getNodeView()
   */
  public function getNodeView() {
    $content = parent::getNodeView();

    if (!$this->viewCanRevealCorrect()) {
      $content['answers'] = array(
        '#type' => 'markup',
        '#value' => '<div class="quiz-members-hidden">Answer hidden</div>',
        '#weight' => 2,
      );
      return $content;
    }

    if (empty($this->node->groups)) {
      $content['nogroups'] = array(
        '#type' => 'markup',
        '#value' => t('No groups have been set for this Grouping question.'),
      );
      return $content;
    }

    $i = 1;
    while (isset($this->node->groups[$i])) {
      $short = $this->node->groups[$i];
      if (drupal_strlen(check_plain($short['name'])) > 0) {
        $data[$i]['name'] = check_plain($short['name']);
        $data[$i]['correct_answers'] = $short['members'];
        $data[$i]['feedback'] = check_markup($short['feedback']);
      }
      $i++;
    }

    $general_feedback = check_markup($this->node->feedback, $this->node->feedback_format);

    // Return themed output.
    $content['answers'] = array(
      '#type' => 'markup',
      '#value' => theme('grouping_question_response', $data, FALSE, $general_feedback, TRUE, TRUE),
    );
    return $content;
  }

  /**
   * Implements getAnsweringForm().
   * @see QuizQuestion#getAnsweringForm($form_state, $rid)
   */
  public function getAnsweringForm(array $form_state, $rid) {
    $form = parent::getAnsweringForm($form_state, $rid);
    $form['#theme'] = 'grouping_question_answering_form';

    if (isset($rid)) {
      // This question has already been memberised. We load the members.
      $response = new GroupingResponse($rid, $this->node);
      $response = $response->getResponse();
    }

    $members = array();

    foreach ($this->node->groups as $group) {
      $new_members = explode(',', $group['members']);
      foreach ($new_members as $id => $new_member) {
        $new_member = trim($new_member);
        if (strlen($new_member) < 1) {
          unset($new_members[$id]);
        }
      }
      $members = array_merge($members, $new_members);
    }
    shuffle($members);

    $form['tries'] = array(
      '#type' => 'fieldset',
      '#title' => '',
      '#collapsible' => FALSE,
      '#tree' => TRUE,
    );

    $form['tries']['#theme'] = 'grouping_question_answer_form_groups';

    $form['tries']['members'] = array(
      '#type' => 'value',
      '#value' => $members,
    );

    for ($i = 1; isset($this->node->groups[$i]); $i++) {
      $short = $this->node->groups[$i];

      $form['tries'][$i]['id'] = array(
        '#type' => 'value',
        '#value' => $short['id'],
      );

      $form['tries'][$i]['user_answers'] = array(
        '#title' => $short['name'],
        '#type' => 'hidden',
        '#attributes' => array('class' => 'quiz-grouping-answer-field'),
        '#default_value' => (isset($response) ? $response[$i]['user_answer'] : ''),
        '#required' => FALSE,
      );
    }
    return $form;
  }

  /**
   * Implements getBodyFieldTitle().
   * @see QuizQuestion#getBodyFieldTitle()
   */
  public function getBodyFieldTitle() {
    return t('Instructions');
  }

  /**
   * Implements getCreationForm().
   * @see QuizQuestion#getCreationForm($form_state)
   */
  public function getCreationForm(array $form_state = NULL) {

    $form['groups'] = array(
      '#type' => 'fieldset',
      '#title' => t('Groups'),
      '#collapsible' => TRUE,
      '#collapsed' => FALSE,
      '#weight' => -4,
      '#tree' => TRUE,
    );

    $form['groups']['#theme'] = 'grouping_question_creation';

    for ($i = 1; $i <= 10; $i++) {

      $short = isset($this->node->groups[$i]) ? $this->node->groups[$i] : NULL;

      // We add id to be able to update the correct alternatives if the node is
      // updated, without destroying existing members reports.
      $form['groups'][$i]['id'] = array(
        '#type' => 'value',
        '#value' => $short['id'],
      );

      $form['groups'][$i]['name'] = array(
        '#type' => 'textfield',
        '#default_value' => isset($short['name']) ? $short['name'] : '',
        '#required' => FALSE,
        '#size' => 100,
        '#maxlength' => 128,
      );

      $form['groups'][$i]['members'] = array(
        '#type' => 'textarea',
        '#default_value' => isset($short['members']) ? $short['members'] : '',
        '#required' => FALSE,
      );

      $form['groups'][$i]['feedback'] = array(
        '#type' => 'textarea',
        '#required' => FALSE,
        '#default_value' => isset($short['feedback']) ? $short['feedback'] : '',
      );
    }

    $form['feedback'] = array(
      '#title' => t('General feedback'),
      '#description' => t('Enter the general feedback the users will see on the report at the end of the quiz.'),
      '#type' => 'textarea',
      '#required' => FALSE,
      '#default_value' => isset($this->node->feedback) ? $this->node->feedback : '',
    );

    $form['feedback_format'] = filter_form($this->node->feedback_format, NULL, array('feedback_format'));

    return $form;
  }

  /**
   * Implements getMaximumScore().
   * @see QuizQuestion#getMaximumScore()
   */
  public function getMaximumScore() {
    $max = 0;
    for ($i = 1; isset($this->node->groups[$i]); $i++) {
      $short = $this->node->groups[$i];
      if (drupal_strlen($short['name']) > 0) {
        $members = explode(',', $short['members']);
        $max += max(count($members), 1);
      }
    }
    return max($max, 1);
  }

}

/**
 * Extension of QuizQuestionResponse
 */
class GroupingResponse extends QuizQuestionResponse {

  protected $user_answers;

  /**
   * Constructor
   */
  public function __construct($result_id, stdClass $question_node, $tries = NULL) {
    parent::__construct($result_id, $question_node, $tries);
    $this->user_answers = array();

    if (is_array($tries)) {
      $responses = isset($tries['answer_id']) ? $tries['answer'] : $tries;
      foreach ($responses as $id => $response) {
        $this->user_answers[$id]['user_answer'] = _grouping_question_clean_csv($response['user_answers']);
        $this->user_answers[$id]['group_id'] = $question_node->groups[$id]['id'];
      }
    }
    else {
      $sql = 'SELECT group_id, user_answer
        FROM {quiz_grouping_question_user_answers} ua
        LEFT OUTER JOIN {quiz_grouping_question_user_answer_multi} uam ON (uam.user_answer_id = ua.id)
        WHERE ua.result_id = %d AND ua.question_nid = %d AND ua.question_vid = %d';
      $res = db_query($sql, $result_id, $this->question->nid, $this->question->vid);
      $i = 1;
      while ($res_o = db_fetch_array($res)) {
        $this->user_answers[$i] = $res_o;
        $i++;
      }
    }
  }

  /**
   * Implements isValid
   *
   * @see QuizQuestionResponse#isValid()
   */
  public function isValid() {
    $answer_counts = array();
    foreach ($this->user_answers as $id => $answers) {
      $user_answers = explode(',', $answers['user_answer']);
      foreach ($user_answers as $answer) {
        if (!isset($answer_counts[$answer])) {
          $answer_counts[$answer] = array('user_count' => 0, 'correct_count' => 0);
        }
        $answer_counts[$answer]['user_count']++;
      }
    }
    foreach ($this->question->groups as $group) {
      $correct_answers = explode(',', $group['members']);
      foreach ($correct_answers as $correct_answer) {
        if (!isset($answer_counts[$correct_answer])) {
          $answer_counts[$correct_answer] = array('user_count' => 0, 'correct_count' => 0);
        }
        $answer_counts[$correct_answer]['correct_count']++;
      }
    }

    foreach ($answer_counts as $answer => $answer_count) {
      if ($answer_count['correct_count'] > 0 && $answer_count['user_count'] > $answer_count['correct_count']) {
        // If an answer is an answer but has been used more times than it
        // appears in groups, the user must have entered the same element into
        // more than one group. This stops them putting every answer in
        // every group.
        return t('Too many "' . $answer . '" entered');
      }
    }

    return TRUE;
  }

  /**
   * Implements save
   *
   * @see QuizQuestionResponse#save()
   */
  public function save() {
    $sql = "INSERT INTO {quiz_grouping_question_user_answers} (question_nid, question_vid, result_id) VALUES (%d, %d, %d)";
    db_query($sql, $this->question->nid, $this->question->vid, $this->rid);
    $user_answer_id = db_last_insert_id('{quiz_grouping_question_user_answers}', 'id');
    for ($i = 1; $i <= count($this->user_answers); $i++) {
      $sql = 'INSERT INTO {quiz_grouping_question_user_answer_multi}
        (user_answer_id, group_id, user_answer)
        VALUES(%d, %d, \'%s\')';
      db_query($sql, $user_answer_id, $this->user_answers[$i]['group_id'], $this->user_answers[$i]['user_answer']);
    }
  }

  /**
   * Implements delete
   *
   * @see QuizQuestionResponse#delete()
   */
  public function delete() {
    $sql = 'DELETE FROM {quiz_grouping_question_user_answer_multi}
      USING {quiz_grouping_question_user_answer_multi}
      INNER JOIN {quiz_grouping_question_user_answers}
      ON {quiz_grouping_question_user_answer_multi}.user_answer_id = {quiz_grouping_question_user_answers}.id
      WHERE {quiz_grouping_question_user_answers}.question_nid = %d
      AND {quiz_grouping_question_user_answers}.question_vid = %d
      AND {quiz_grouping_question_user_answers}.result_id = %d';
    db_query($sql, $this->question->nid, $this->question->vid, $this->rid);
    $sql = 'DELETE FROM {quiz_grouping_question_user_answers}
    WHERE result_id = %d AND question_nid = %d AND question_vid = %d';
    db_query($sql, $this->rid, $this->question->nid, $this->question->vid);
  }

  /**
   * Implements score
   *
   * @see QuizQuestionResponse#score()
   */
  public function score() {
    $score = 0;
    if ($this->isValid() !== TRUE) {
      return 0;
    }

    foreach ($this->question->groups as $key => $group) {
      $correct_answers = explode(',', $group['members']);
      $user_answers = explode(',', $this->user_answers[$key]['user_answer']);
      foreach ($user_answers as $user_answer) {
        if (in_array($user_answer, $correct_answers)) {
          $score++;
        }
      }
    }
    return max($score, 0);
  }

  /**
   * Implements getResponse
   *
   * @see QuizQuestionResponse#getResponse()
   */
  public function getResponse() {
    return $this->user_answers;
  }

  /**
   * Implements getReportFormResponse
   * @see QuizQuestionResponse#getReportFormResponse
   */
  public function getReportFormResponse($showpoints = TRUE, $showfeedback = TRUE, $allow_scoring = FALSE) {
    if (empty($this->question->groups)) {
      return array(
        '#type' => 'markup',
        '#value' => t('No groups have been set for this Grouping question.'),
      );
    }

    $data = array();

    $i = 1;
    while (isset($this->question->groups[$i])) {
      $short = $this->question->groups[$i];
      if (drupal_strlen(check_plain($short['name'])) > 0) {
        $data[$i]['name'] = check_plain($short['name']);
        $data[$i]['correct_answers'] = $short['members'];
        $data[$i]['user_answers'] = $this->user_answers[$i]['user_answer'];
        $data[$i]['feedback'] = check_markup($short['feedback']);
      }
      $i++;
    }

    $general_feedback = check_markup($this->question->feedback, $this->question->feedback_format);

    // Return themed output.
    return array(
      '#type' => 'markup',
      '#value' => theme('grouping_question_response', $data, TRUE, $general_feedback, $showpoints, $showfeedback),
    );
  }

}
